
#include <stdio.h>
#include <stdlib.h>
#include "cell.h"
#include "rif.h"
#include "csi.h"

extern void	Error(char *msg),
		RangeList(char *str,size_t lim,char *pv);

#define U_C	unsigned char
#define U_S	unsigned short
#define U_L	unsigned long

static size_t	FWriteReverse(void *P,size_t S,size_t C,FILE *F)
{
	U_C	*p=(U_C *)P;
	size_t	c,s;

	for (--p, c=0; c<C; c++, p+=S)
	for (p+=S, s=0; s<S; s++, p--) if (fputc(*p,F)!=(*p)) return c;

	return C;
}

static U_S	flag=0;
static FILE	*file;
static size_t	(*FWrite)();
static int	X,Y,Z;

#ifdef	DEBUG
static U_L	stl_facets=0;
#endif

static void	Facet(int nx,int ny,int nz,int x0,int y0,int z0,
					   int x1,int y1,int z1,
					   int x2,int y2,int z2)
{
	float	f[12];

	f[0]=(float)   nx ; f[ 1]=(float)   ny ; f[ 2]=(float)   nz ;
	f[3]=(float)(X+x0); f[ 4]=(float)(Y+y0); f[ 5]=(float)(Z+z0);
	f[6]=(float)(X+x1); f[ 7]=(float)(Y+y1); f[ 8]=(float)(Z+z1);
	f[9]=(float)(X+x2); f[10]=(float)(Y+y2); f[11]=(float)(Z+z2);

	if ((*FWrite)((U_C *)f,sizeof(float),12,file)!=12 ||
	    (*FWrite)((U_C *)&flag,sizeof(U_S),1,file)!=1)
	    Error("STL facet data write error.");

#ifdef	DEBUG
	++stl_facets;
#endif
}

typedef struct List {
		U_S		x,y;
		struct List	*next;
	} List;

static void	MinusZ(List *a,List *b,List *c)
{
	Facet(0,0,-1,c->x,c->y,0,b->x,b->y,0,a->x,a->y,0);
}

static void	PlusZ(List *a,List *b,List *c)
{
	Facet(0,0,+1,a->x,a->y,1,b->x,b->y,1,c->x,c->y,1);
}

static void	MinusX(List *a,List *b,List *c)
{
	Facet(-1,0,0,0,c->x,c->y,0,b->x,b->y,0,a->x,a->y);
}

static void	PlusX(List *a,List *b,List *c)
{
	Facet(+1,0,0,1,a->x,a->y,1,b->x,b->y,1,c->x,c->y);
}

static void	MinusY(List *a,List *b,List *c)
{
	Facet(0,-1,0,c->y,0,c->x,b->y,0,b->x,a->y,0,a->x);
}

static void	PlusY(List *a,List *b,List *c)
{
	Facet(0,+1,0,a->y,1,a->x,b->y,1,b->x,c->y,1,c->x);
}

static void	(*Print)();

static void	Flush(List *list)
{
	U_L	idx,cnt=0;
	List	**array,**head,**tail,*oList,*nList,*stack=list;
	int	oSide,nSide,x0,y0,x1,y1,x2,y2;

	while ((list=list->next)!=NULL) ++cnt;

	if ((array=(List **)malloc(sizeof(List *)*cnt))==NULL)
	    Error("array memory allocation error.");

	list=stack;
	for (idx=0; idx<cnt; idx++) {
	    array[idx]=list; list=list->next;
	}
	free(list);

	stack->next=NULL;

	head=array+1; tail=array+cnt-1;
	if ((*head)->y>(*tail)->y) {
	    oSide=0; oList=(*(head++));
	}
	else {
	    oSide=1; oList=(*(tail--));
	}
	while (head!=tail) {
	    if ((*head)->y>(*tail)->y) {
		nSide=0; nList=(*(head++));
	    }
	    else {
		nSide=1; nList=(*(tail--));
	    }
	    if (oSide==nSide) {
		x2=oList->x-(x0=nList->x); y2=oList->y-(y0=nList->y);
		if (oSide==0)
		    do {
			x1=x2; y1=y2;
			if (y1*(x2=stack->x-x0)<=x1*(y2=stack->y-y0)) break;

			(*Print)(stack,oList,nList); oList=stack;
		    } while ((stack=stack->next)!=NULL);
		else /* if (oSide==1) */
		    do {
			x1=x2; y1=y2;
			if (y1*(x2=stack->x-x0)>=x1*(y2=stack->y-y0)) break;

			(*Print)(stack,nList,oList); oList=stack;
		    } while ((stack=stack->next)!=NULL);
	    }
	    else {
		list=oList;
		if (oSide==0)
		    do {
			(*Print)(stack,oList,nList); oList=stack;
		    } while ((stack=stack->next)!=NULL);
		else /* if (oSide==1) */
		    do {
			(*Print)(stack,nList,oList); oList=stack;
		    } while ((stack=stack->next)!=NULL);

		oList=list;
	    }
	    oList->next=stack; stack=oList; oList=nList; oSide=nSide;
	}
	nList=(*head);
	if (oSide==0)
	    do {
		(*Print)(stack,oList,nList); oList=stack;
	    } while ((stack=stack->next)!=NULL);
	else /* if (oSide==1) */
	    do {
		(*Print)(stack,nList,oList); oList=stack;
	    } while ((stack=stack->next)!=NULL);

	for (idx=0; idx<cnt; idx++) free(array[idx]); free(array);
}

static List	*Alloc(int x,int y,List *next)
{
	List	*list=(List *)malloc(sizeof(*list));

	if (list==NULL) Error("list memory allocation error.");

	list->x=(U_S)x; list->y=(U_S)y; list->next=next; return list;
}

static List	*hori,**vert,**help,**ptr;

static void	Corner(int x,int y,int a,int b,int c,int d)
{
	List	*temp,*next;

#define VERT	vert[x]
#define HELP	help[x]

	switch(((a==1)?8:0)|((b==1)?4:0)|((c==1)?2:0)|((d==1)?1:0)) {
	    case 0x1:
		VERT=Alloc(x,y+1,Alloc(x,y,hori=Alloc(x+1,y,NULL)));
		ptr=(&HELP);
		break;
	    case 0x2:
		VERT->y=(U_S)y;
		next=HELP->next;
		if (HELP->x==next->x && HELP->y==next->y) {
		    hori=Alloc(x+1,y,Alloc(x,y,next));
		    HELP->next=Alloc(x,y,NULL); Flush(VERT);
		}
		else
		    hori=Alloc(x+1,y,VERT);
		break;
	    case 0x3:
		if (a!=b) {
		    VERT->y=(U_S)y; VERT=Alloc(x,y+1,VERT);
		}
		ptr=(&HELP);
		break;
	    case 0x4:
		hori->x=(U_S)x; hori->next=VERT=Alloc(x,y+1,NULL);
		*ptr=hori;
		break;
	    case 0x5:
		if (a!=c) {
		    hori->x=(U_S)x; hori=hori->next=Alloc(x+1,y,NULL);
		}
		break;
	    case 0x6:
		VERT->y=(U_S)y;
		next=HELP->next;
		if (HELP->x==next->x && HELP->y==next->y) {
		    temp=Alloc(x+1,y,Alloc(x,y,next));
		    HELP->next=Alloc(x,y,NULL); Flush(VERT);
		}
		else
		    temp=Alloc(x+1,y,VERT);
		hori->x=(U_S)x; hori->next=VERT=Alloc(x,y+1,NULL);
		*ptr=hori;
		hori=temp;
		break;
	    case 0x7:
		hori->x=(U_S)x; VERT->y=(U_S)y;
		next=HELP->next;
		if (HELP->x==next->x && HELP->y==next->y) {
		    hori->next=Alloc(x,y,next);
		    HELP->next=Alloc(x,y,NULL); Flush(VERT);
		}
		else
		    hori->next=VERT;
		*ptr=hori;
		break;
	    case 0x8:
		VERT->y=(U_S)y; hori->x=(U_S)x; Flush(hori);
		break;
	    case 0x9:
		VERT->y=(U_S)y; hori->x=(U_S)x; Flush(hori);
		VERT=Alloc(x,y+1,Alloc(x,y,hori=Alloc(x+1,y,NULL)));
		ptr=(&HELP);
		break;
	    case 0xa:
		if (b!=d) {
		    hori->x=(U_S)x; hori=Alloc(x+1,y,hori);
		}
		break;
	    case 0xb:
		hori->x=(U_S)x; VERT=Alloc(x,y+1,hori);
		ptr=(&HELP); HELP=hori;
		break;
	    case 0xc:
		if (c!=d) {
		    VERT->y=(U_S)y; VERT=VERT->next=Alloc(x,y+1,NULL);
		}
		break;
	    case 0xd:
		VERT->y=(U_S)y;
		temp=(*ptr); next=temp->next;
		if (temp->x==next->x && temp->y==next->y) {
		    Flush(Alloc(x,y,next));
		    temp->next=Alloc(x,y,hori=Alloc(x+1,y,NULL));
		}
		else
		    VERT->next=hori=Alloc(x+1,y,NULL);
		break;
	    case 0xe:
		temp=(*ptr); next=temp->next;
		if (temp->x==next->x && temp->y==next->y)
		    hori=Alloc(x+1,y,Alloc(x,y,next));
		else
		    hori=Alloc(x+1,y,Alloc(x,y,Alloc(temp->x,temp->y,next)));
		*ptr=temp->next=Alloc(x,y,VERT=Alloc(x,y+1,NULL));
		break;
	}
}

#define LIM	(size_t)((Cell)(~0))

int	main(int argc,char **argv)
{
	char	**path,pv[LIM+1],**LU;
	int	Nx,Ny,Nz,n,y,x1,y1,z1,x2,y2,z2,z,nx,ny,x,nz,N,x0,y0,z0;
	U_C	***image,**slice,mask;
	Cell	**cell;
	U_L	facets,pixels=0,Vt=0,Vr=0,Vs=0,faces=0;

	if (argc!=5 && argc!=6 && argc!=8 && argc!=9)
	Error("usage : si_stl_B directory nameFile rangeList {R G B} STL\n"
	      "        si_stl_B directory nameFile lower upper {R G B} STL");

	path=CheckSliceImages(argv[1],argv[2],&Nx,&Ny,&Nz,NULL);
{
	U_L	low,upp,idx;

	if (argc==5 || argc==8)
	    RangeList(argv[3],LIM,pv);
	else {
	    if (sscanf(argv[3],"%lu",&low)!=1 ||
		sscanf(argv[4],"%lu",&upp)!=1 || low>upp)
		Error("bad pixel value range.");

	    for (idx=0; idx<=LIM; idx++) pv[idx]=(idx>=low && idx<=upp);
	}
}{
	U_S	R,G,B;

	if (argc==8 || argc==9) {
	    if (sscanf(argv[argc-4],"%hu",&R)!=1 || R>=256 ||
		sscanf(argv[argc-3],"%hu",&G)!=1 || G>=256 ||
		sscanf(argv[argc-2],"%hu",&B)!=1 || B>=256)
		Error("bad RGB color component value(s).");

	    flag=32768|((R&248)<<7)|((G&248)<<2)|((B&248)>>3);
	}
}
	(void)fprintf(stderr,"%d\t%d\t%d\n",Nx,Ny,Nz);

	n=(Nz+7)/8;
	if ((image=(U_C ***)malloc(sizeof(U_C **)*n))==NULL ||
	    (*image=(U_C **)malloc(sizeof(U_C *)*n*Ny))==NULL ||
	    (**image=(U_C *)calloc((size_t)n*Ny*Nx,sizeof(U_C)))==NULL ||
	    (LU=(char **)malloc(sizeof(char *)*(Ny+2)))==NULL ||
	    (*LU=(char *)calloc((size_t)(Ny+2)*(Nx+2),sizeof(char)))==NULL)
	    Error("image memory allocation error.");

	++LU; ++LU[-1]; for (y=0; y<=Ny; y++) LU[y]=LU[y-1]+Nx+2;

	x1=Nx-1; y1=Ny-1; z1=Nz-1; x2=y2=z2=0;
	for (z=0; z<=Nz; z++) {
	    if (z<Nz) {
		ReadImageFile(path[z],&nx,&ny,NULL,&cell,NULL); free(path[z]);

		if (z%8==0) {
		    if (z==0) {
			slice=image[0]; y=1;
		    }
		    else {
			n=z/8; slice=image[n]=image[n-1]+Ny; y=0;
		    }
		    for (; y<Ny; y++) slice[y]=slice[y-1]+Nx; mask=1;
		}
		else
		    mask<<=1;

		for (y=0; y<Ny; y++) {
		    for (x=0; x<Nx; x++)
			if (pv[(x<nx && y<ny)?cell[y][x]:0]) {
			    slice[y][x]|=mask;
			    if (x<x1) x1=x; if (x>x2) x2=x;
			    if (y<y1) y1=y; if (y>y2) y2=y;
			    if (z<z1) z1=z; if (z>z2) z2=z; ++pixels;

			    LU[y][x]=((LU[y][x]<<1)|1)&3;
			}
			else
			    LU[y][x]=(LU[y][x]<<1)&2;

		    if (y<ny) free(cell[y]);
		}
		free(cell);
	    }
	    else {
		free(path);

		for (y=0; y<Ny; y++)
		for (x=0; x<Nx; x++) LU[y][x]=(LU[y][x]<<1)&2;
            }
	    for (y=0; y<=Ny; y++)
	    for (x=0; x<=Nx; x++)
{
	int	lnw=LU[y-1][x-1]&1,unw=LU[y-1][x-1]>>1,
		lne=LU[y-1][x  ]&1,une=LU[y-1][x  ]>>1,
		lsw=LU[y  ][x-1]&1,usw=LU[y  ][x-1]>>1,
		lse=LU[y  ][x  ]&1,use=LU[y  ][x  ]>>1,
		fx=(lnw!=lne)+(lsw!=lse)+(unw!=une)+(usw!=use),
		fy=(lnw!=lsw)+(lne!=lse)+(unw!=usw)+(une!=use),
		fz=(lnw!=unw)+(lne!=une)+(lsw!=usw)+(lse!=use),f=fx+fy+fz;

	if (f) {
	    if (fx && fy && fz) ++Vt; else if (f==8) ++Vr;

	    ++Vs; faces+=f;
	}
}
	}
	free(LU[-1]-1); free(LU-1);

	if (pixels==0) Error("no object.");

	faces/=4; facets=2*(Vt-Vr-Vs+faces);
	(void)fprintf(stderr,"%lu\t%d\t%d\t%d\t%d\t%d\t%d\n%lu\t%lu\n",
			     pixels,x1,y1,z1,x2,y2,z2,faces,facets);

	nx=x2-x1+1;
	ny=y2-y1+1;
	nz=z2-z1+1; N=(nx>ny)?(nx>nz)?nx:nz:(ny>nz)?ny:nz;

	n=N+1;
	if ((LU=(char **)malloc(sizeof(char *)*n))==NULL ||
	    (*LU=(char *)calloc((size_t)n*n,sizeof(char)))==NULL ||
	    (vert=(List **)malloc(sizeof(List *)*n))==NULL ||
	    (help=(List **)malloc(sizeof(List *)*N))==NULL)
	    Error("work memory allocation error.");

	++LU; ++LU[-1]; for (n=0; n<N; n++) LU[n]=LU[n-1]+N+1;

	if (argv[argc-1][0]=='-')
	    file=stdout;
	else
	    if ((file=fopen(argv[argc-1],"wb"))==NULL)
		Error("STL file not open.");

	for (n=0; n<80; n++)
	    if (fputc(0,file)!=0) Error("STL header write error.");
{
	U_S		u_s=0x0123;
	U_C		*u_c=(U_C *)&u_s;
	unsigned	u=(unsigned)facets;

	FWrite=(*u_c==(u_s&255))?(size_t (*)())fwrite:FWriteReverse;

	if ((*FWrite)((U_C *)&u,sizeof(unsigned),1,file)!=1)
	    Error("STL facet number write error.");
}
#define CORNER(h,v)	Corner(h,v,LU[v-1][h-1],LU[v][h-1],LU[v-1][h],LU[v][h])
#define CORNER_H(h,v)	Corner(h,v,LU[v-1][h-1],LU[v][h-1],0,0)
#define CORNER_V(h,v)	Corner(h,v,LU[v-1][h-1],0,LU[v-1][h],0)
#define CORNER_HV(h,v)	Corner(h,v,LU[v-1][h-1],0,0,0)

X=x1; Y=y1;

    Print=(void (*)())MinusZ;
/*
	for (y=0; y<ny; y++) for (x=0; x<nx; x++) LU[y][x]=0;
*/
	for (Z=z1; Z<=z2; Z++) {
	    slice=image[Z/8]; mask=(U_C)(1<<(Z%8));
	    for (y0=y1, y=0; y<ny; y++, y0++) {
		for (x0=x1, x=0; x<nx; x++, x0++) {
		    LU[y][x]=((LU[y][x]<<1)|((slice[y0][x0]&mask)!=0))&3;

		    CORNER(x,y);
		}
		CORNER_H(x,y);
	    }
	    for (x=0; x<nx; x++) CORNER_V(x,y);

	    CORNER_HV(x,y);
	}

    Print=(void (*)())PlusZ;

	for (y=0; y<ny; y++) for (x=0; x<nx; x++) LU[y][x]=0;

	for (Z=z2; Z>=z1; Z--) {
	    slice=image[Z/8]; mask=(U_C)(1<<(Z%8));
	    for (y0=y1, y=0; y<ny; y++, y0++) {
		for (x0=x1, x=0; x<nx; x++, x0++) {
		    LU[y][x]=((LU[y][x]<<1)|((slice[y0][x0]&mask)!=0))&3;

		    CORNER(x,y);
		}
		CORNER_H(x,y);
	    }
	    for (x=0; x<nx; x++) CORNER_V(x,y);

	    CORNER_HV(x,y);
	}

Y=y1; Z=z1;

    Print=(void (*)())MinusX;

	for (z=0; z<nz; z++) for (y=0; y<ny; y++) LU[z][y]=0;

	for (X=x1; X<=x2; X++) {
	    for (z0=z1, z=0; z<nz; z++, z0++) {
		slice=image[z0/8]; mask=(U_C)(1<<(z0%8));
		for (y0=y1, y=0; y<ny; y++, y0++) {
		    LU[z][y]=((LU[z][y]<<1)|((slice[y0][X]&mask)!=0))&3;

		    CORNER(y,z);
		}
		CORNER_H(y,z);
	    }
	    for (y=0; y<ny; y++) CORNER_V(y,z);

	    CORNER_HV(y,z);
	}

    Print=(void (*)())PlusX;

	for (z=0; z<nz; z++) for (y=0; y<ny; y++) LU[z][y]=0;

	for (X=x2; X>=x1; X--) {
	    for (z0=z1, z=0; z<nz; z++, z0++) {
		slice=image[z0/8]; mask=(U_C)(1<<(z0%8));
		for (y0=y1, y=0; y<ny; y++, y0++) {
		    LU[z][y]=((LU[z][y]<<1)|((slice[y0][X]&mask)!=0))&3;

		    CORNER(y,z);
		}
		CORNER_H(y,z);
	    }
	    for (y=0; y<ny; y++) CORNER_V(y,z);

	    CORNER_HV(y,z);
	}

Z=z1; X=x1;

    Print=(void (*)())MinusY;

	for (x=0; x<nx; x++) for (z=0; z<nz; z++) LU[x][z]=0;

	for (Y=y1; Y<=y2; Y++) {
	    for (x0=x1, x=0; x<nx; x++, x0++) {
		for (z0=z1, z=0; z<nz; z++, z0++) {
	    LU[x][z]=((LU[x][z]<<1)|((image[z0/8][Y][x0]&(1<<(z0%8)))!=0))&3;

		    CORNER(z,x);
		}
		CORNER_H(z,x);
	    }
	    for (z=0; z<nz; z++) CORNER_V(z,x);

	    CORNER_HV(z,x);
	}

    Print=(void (*)())PlusY;

	for (x=0; x<nx; x++) for (z=0; z<nz; z++) LU[x][z]=0;

	for (Y=y2; Y>=y1; Y--) {
	    for (x0=x1, x=0; x<nx; x++, x0++) {
		for (z0=z1, z=0; z<nz; z++, z0++) {
	    LU[x][z]=((LU[x][z]<<1)|((image[z0/8][Y][x0]&(1<<(z0%8)))!=0))&3;

		    CORNER(z,x);
		}
		CORNER_H(z,x);
	    }
	    for (z=0; z<nz; z++) CORNER_V(z,x);

	    CORNER_HV(z,x);
	}
	if (file!=stdout) (void)fclose(file);

#ifdef	DEBUG
	(void)fprintf(stderr,"%lu\n",stl_facets);
#endif

	free(help); free(vert); free(LU[-1]-1); free(LU-1);
	free(**image); free(*image); free(image);

	return 0;
}
