
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include "cell.h"
#include "rif.h"
#include "csi.h"

extern void	Error(char *msg);

#define U_C	unsigned char
#define U_S	unsigned short
#define U_L	unsigned long

#define RL_T	U_S
#include "rl.h"

static size_t	FWriteReverse(void *P,size_t S,size_t C,FILE *F)
{
	U_C	*p=(U_C *)P;
	size_t	c,s;

	for (--p, c=0; c<C; c++, p+=S)
	for (p+=S, s=0; s<S; s++, p--) if (fputc(*p,F)!=(*p)) return c;

	return C;
}

static FILE	*file;
static size_t	(*FWrite)();
static int	X,Y,Z;

#ifdef	DEBUG
static U_L	stl_facets=0;
#endif

static void	Facet(int nx,int ny,int nz,int x0,int y0,int z0,
					   int x1,int y1,int z1,
					   int x2,int y2,int z2,U_S flag)
{
	float	f[12];

	f[0]=(float)   nx ; f[ 1]=(float)   ny ; f[ 2]=(float)   nz ;
	f[3]=(float)(X+x0); f[ 4]=(float)(Y+y0); f[ 5]=(float)(Z+z0);
	f[6]=(float)(X+x1); f[ 7]=(float)(Y+y1); f[ 8]=(float)(Z+z1);
	f[9]=(float)(X+x2); f[10]=(float)(Y+y2); f[11]=(float)(Z+z2);

	if ((*FWrite)((U_C *)f,sizeof(float),12,file)!=12 ||
	    (*FWrite)((U_C *)&flag,sizeof(U_S),1,file)!=1)
	    Error("STL facet data write error.");

#ifdef	DEBUG
	++stl_facets;
#endif
}

typedef struct List {
		U_S		x,y;
		struct List	*next;
	} List;

static void	MinusZ(List *a,List *b,List *c,U_S flag)
{
	Facet(0,0,-1,c->x/2,c->y/2,0,b->x/2,b->y/2,0,a->x/2,a->y/2,0,flag);
}

static void	PlusZ(List *a,List *b,List *c,U_S flag)
{
	Facet(0,0,+1,a->x/2,a->y/2,1,b->x/2,b->y/2,1,c->x/2,c->y/2,1,flag);
}

static void	MinusX(List *a,List *b,List *c,U_S flag)
{
	Facet(-1,0,0,0,c->x/2,c->y/2,0,b->x/2,b->y/2,0,a->x/2,a->y/2,flag);
}

static void	PlusX(List *a,List *b,List *c,U_S flag)
{
	Facet(+1,0,0,1,a->x/2,a->y/2,1,b->x/2,b->y/2,1,c->x/2,c->y/2,flag);
}

static void	MinusY(List *a,List *b,List *c,U_S flag)
{
	Facet(0,-1,0,c->y/2,0,c->x/2,b->y/2,0,b->x/2,a->y/2,0,a->x/2,flag);
}

static void	PlusY(List *a,List *b,List *c,U_S flag)
{
	Facet(0,+1,0,a->y/2,1,a->x/2,b->y/2,1,b->x/2,c->y/2,1,c->x/2,flag);
}

static void	(*Print)();

static void	Flush(List *list,U_S flag)
{
	U_L	idx,cnt=0;
	List	**array,**head,**tail,*oList,*nList,*stack=list;
	int	oSide,nSide,x0,y0,x1,y1,x2,y2;

	while ((list=list->next)!=NULL) ++cnt;

	if ((array=(List **)malloc(sizeof(List *)*cnt))==NULL)
	    Error("array memory allocation error.");

	list=stack;
	for (idx=0; idx<cnt; idx++) {
	    array[idx]=list; list=list->next;
	}
	free(list);

	stack->next=NULL;

	head=array+1; tail=array+cnt-1;
	if ((*head)->y>(*tail)->y) {
	    oSide=0; oList=(*(head++));
	}
	else {
	    oSide=1; oList=(*(tail--));
	}
	while (head!=tail) {
	    if ((*head)->y>(*tail)->y) {
		nSide=0; nList=(*(head++));
	    }
	    else {
		nSide=1; nList=(*(tail--));
	    }
	    if (oSide==nSide) {
		x2=oList->x-(x0=nList->x); y2=oList->y-(y0=nList->y);
		if (oSide==0)
		    do {
			x1=x2; y1=y2;
			if (y1*(x2=stack->x-x0)<=x1*(y2=stack->y-y0)) break;

			(*Print)(stack,oList,nList,flag); oList=stack;
		    } while ((stack=stack->next)!=NULL);
		else /* if (oSide==1) */
		    do {
			x1=x2; y1=y2;
			if (y1*(x2=stack->x-x0)>=x1*(y2=stack->y-y0)) break;

			(*Print)(stack,nList,oList,flag); oList=stack;
		    } while ((stack=stack->next)!=NULL);
	    }
	    else {
		list=oList;
		if (oSide==0)
		    do {
			(*Print)(stack,oList,nList,flag); oList=stack;
		    } while ((stack=stack->next)!=NULL);
		else /* if (oSide==1) */
		    do {
			(*Print)(stack,nList,oList,flag); oList=stack;
		    } while ((stack=stack->next)!=NULL);

		oList=list;
	    }
	    oList->next=stack; stack=oList; oList=nList; oSide=nSide;
	}
	nList=(*head);
	if (oSide==0)
	    do {
		(*Print)(stack,oList,nList,flag); oList=stack;
	    } while ((stack=stack->next)!=NULL);
	else /* if (oSide==1) */
	    do {
		(*Print)(stack,nList,oList,flag); oList=stack;
	    } while ((stack=stack->next)!=NULL);

	for (idx=0; idx<cnt; idx++) free(array[idx]); free(array);
}

static List	*Alloc(int x,int y,List *next)
{
	List	*list=(List *)malloc(sizeof(*list));

	if (list==NULL) Error("list memory allocation error.");

	list->x=(U_S)x; list->y=(U_S)y; list->next=next; return list;
}

static List	*hori,**vert,**help,**ptr;

static void	Corner(int x,int y,U_S a,U_S b,U_S c,U_S d,int h,int v)
{
	List	*temp,*next;

#define VERT	vert[x]
#define HELP	help[x]

	switch(((a!=0)?8:0)|((b!=0)?4:0)|((c!=0)?2:0)|((d!=0)?1:0)) {
	    case 0x1:
		VERT=Alloc(x,y+1,Alloc(x,y,hori=Alloc(x+1,y,NULL)));
		ptr=(&HELP);
		break;
	    case 0x2:
		VERT->y=(U_S)y;
		next=HELP->next;
		if (HELP->x==next->x && HELP->y==next->y) {
		    hori=Alloc(x+1,y,Alloc(x,y,next));
		    HELP->next=Alloc(x,y,NULL); Flush(VERT,c);
		}
		else
		    hori=Alloc(x+1,y,VERT);
		break;
	    case 0x3:
		if (v) {
		    VERT->y=(U_S)y; VERT=Alloc(x,y+1,VERT);
		}
		ptr=(&HELP);
		break;
	    case 0x4:
		hori->x=(U_S)x; hori->next=VERT=Alloc(x,y+1,NULL);
		*ptr=hori;
		break;
	    case 0x5:
		if (h) {
		    hori->x=(U_S)x; hori=hori->next=Alloc(x+1,y,NULL);
		}
		break;
/*
	    case 0x6:
		break;
*/
	    case 0x7:
		hori->x=(U_S)x; VERT->y=(U_S)y;
		next=HELP->next;
		if (HELP->x==next->x && HELP->y==next->y) {
		    hori->next=Alloc(x,y,next);
		    HELP->next=Alloc(x,y,NULL); Flush(VERT,d);
		}
		else
		    hori->next=VERT;
		*ptr=hori;
		break;
	    case 0x8:
		VERT->y=(U_S)y; hori->x=(U_S)x; Flush(hori,a);
		break;
/*
	    case 0x9:
		break;
*/
	    case 0xa:
		if (h) {
		    hori->x=(U_S)x; hori=Alloc(x+1,y,hori);
		}
		break;
	    case 0xb:
		hori->x=(U_S)x; VERT=Alloc(x,y+1,hori);
		ptr=(&HELP); HELP=hori;
		break;
	    case 0xc:
		if (v) {
		    VERT->y=(U_S)y; VERT=VERT->next=Alloc(x,y+1,NULL);
		}
		break;
	    case 0xd:
		VERT->y=(U_S)y;
		temp=(*ptr); next=temp->next;
		if (temp->x==next->x && temp->y==next->y) {
		    Flush(Alloc(x,y,next),b);
		    temp->next=Alloc(x,y,hori=Alloc(x+1,y,NULL));
		}
		else
		    VERT->next=hori=Alloc(x+1,y,NULL);
		break;
	    case 0xe:
		temp=(*ptr); next=temp->next;
		if (temp->x==next->x && temp->y==next->y)
		    hori=Alloc(x+1,y,Alloc(x,y,next));
		else
		    hori=Alloc(x+1,y,Alloc(x,y,Alloc(temp->x,temp->y,next)));
		*ptr=temp->next=Alloc(x,y,VERT=Alloc(x,y+1,NULL));
		break;
	}
}

static char	*AC0,*BD0;
static U_S	*AC1,*BD1,*AC2,*BD2;

static void	Erosion(int Nx,int Ny,U_S **lo,U_S **up)
{
	int	x1,x2,y1,y2,h,v,a,b,c,d;
	U_S	*lAC,*lBD,*uAC,*uBD,A,B,C,D;

	for (x1=x2=0; x1<Nx; x1++, x2+=2) {
	    BD0[x1]=0; BD1[x1]=BD2[x2]=BD2[x2+1]=0;
	}
	for (y1=y2=0; y1<Ny; y1++, y2+=2) {
	    lAC=lo[y1-1]; lBD=lo[y1];
	    uAC=up[y1-1]; uBD=up[y1];

	    for (x1=x2=0; x1<Nx; x1++, x2+=2) {
		AC0[x1]=BD0[x1]; BD0[x1]=(lBD[x1]!=0)|(uBD[x1]!=0)<<1;

		A=AC1[x1-1]; C=AC1[x1]=BD1[x1];
		B=BD1[x1-1]; D=BD1[x1]=(BD0[x1]==1)?lBD[x1]:0;

		if (D!=0 && D==C) {
		    AC2[x2]=(D==A && D==B)?D:0; AC2[x2+1]=D;
		}
		else
		    AC2[x2]=AC2[x2+1]=0;

		if (D!=B && D!=C)
		    h=v=1;
		else
		    if ((d=BD0[x1])==1)
			h=v=0;
		    else {
			b=BD0[x1-1]; c=AC0[x1];
			if (d==0) {
			    h=b==2 || b==3;
			    v=c==2 || c==3;
			}
			else { /* d==2 || d==3 */
			    h=b==0 || b+d==5 || (b==d && uBD[x1-1]!=uBD[x1]);
			    v=c==0 || c+d==5 || (c==d && uAC[x1  ]!=uBD[x1]);
			}
		    }

		Corner(x2,y2,BD2[x2-1],AC2[x2-1],BD2[x2],AC2[x2],h,v);

		Corner(x2+1,y2,BD2[x2],AC2[x2],BD2[x2+1],AC2[x2+1],0,0);
	    }
	    Corner(x2,y2,BD2[x2-1],AC2[x2-1],0,0,0,0);

	    for (x1=x2=0; x1<Nx; x1++, x2+=2) {
		A=AC1[x1-1]; C=AC1[x1];
		B=BD1[x1-1]; D=BD1[x1];

		if (D!=0) {
		    BD2[x2]=(D==B)?D:0; BD2[x2+1]=D;
		}
		else
		    BD2[x2]=BD2[x2+1]=0;

		Corner(x2,y2+1,AC2[x2-1],BD2[x2-1],AC2[x2],BD2[x2],0,0);

		if (A!=B && A!=C)
		    h=v=1;
		else
		    if ((a=AC0[x1-1])==1)
			h=v=0;
		    else {
			c=AC0[x1]; b=BD0[x1-1];
			if (a==0) {
			    h=c==2 || c==3;
			    v=b==2 || b==3;
			}
			else { /* a==2 || a==3 */
			    h=c==0 || c+a==5 || (c==a && uAC[x1  ]!=uAC[x1-1]);
			    v=b==0 || b+a==5 || (b==a && uBD[x1-1]!=uAC[x1-1]);
			}
		    }

		Corner(x2+1,y2+1,AC2[x2],BD2[x2],AC2[x2+1],BD2[x2+1],h,v);
            }
	    Corner(x2,y2+1,AC2[x2-1],BD2[x2-1],0,0,0,0);
	}
	for (x1=x2=0; x1<Nx; x1++, x2+=2) {
	    Corner(x2  ,y2,BD2[x2-1],0,BD2[x2  ],0,0,0);
	    Corner(x2+1,y2,BD2[x2  ],0,BD2[x2+1],0,0,0);
	}
        Corner(x2,y2,BD2[x2-1],0,0,0,0,0);
}

#define LIM	(size_t)((Cell)(~0))
#define LEN	1024

int	main(int argc,char **argv)
{
	char	**path;
	int	Nx,Ny,Nz,y,x1,y1,z1,x2,y2,z2,z,nx,ny,x,nz,N,n,x0,y0,z0;
	U_S	flag[LIM+1],**lo,**up,**lu;
	Cell	***image,**slice,**cell;
	U_L	facets,pixels=0,Vt=0,Vr=0,Vs=0,faces=0;

	if (argc!=4) Error("usage : si_stl_C directory nameFile STL");

	path=CheckSliceImages(argv[1],argv[2],&Nx,&Ny,&Nz,NULL);
{
	U_L	idx,low,upp;
	char	str[LEN],rl[LEN],*bcm="bad color map.";
	U_S	R1,G1,B1,R2,G2,B2,f;
	double	d0,d1,d2;

	flag[0]=0; for (idx=1; idx<=LIM; idx++) flag[idx]=65535;

#define LOW_UPP()	sscanf(rl,"%lu",&low)!=1 || low>upp || upp>LIM
#define FLAG(R,G,B)	\
	32768|(((U_S)(R)&248)<<7)|(((U_S)(G)&248)<<2)|(((U_S)(B)&248)>>3)

	while (fgets(str,LEN,stdin)!=NULL && str[strlen(str)-1]=='\n')
	    switch(sscanf(str,"%s %lu %hu %hu %hu %hu %hu %hu",
			      rl,&upp,&R1,&G1,&B1,&R2,&G2,&B2)) {
		case EOF:
		    continue;
		case 1:
		    RangeList(rl,LIM,flag,0); break;
		case 2:
		    if (LOW_UPP()) Error(bcm);

		    for (idx=low; idx<=upp; idx++) flag[idx]=0;
		    break;
		case 4:
		    if (upp>=256 || R1>=256 || G1>=256) Error(bcm);

		    RangeList(rl,LIM,flag,FLAG(upp,R1,G1));
		    break;
		case 5:
		    if (LOW_UPP() || R1>=256 || G1>=256 || B1>=256) Error(bcm);

		    f=FLAG(R1,G1,B1);
		    for (idx=low; idx<=upp; idx++) flag[idx]=f;
		    break;
		case 8:
		    if (LOW_UPP() || R1>=256 || G1>=256 || B1>=256 ||
				     R2>=256 || B2>=256 || B2>=256) Error(bcm);

		    if (low==upp)
			flag[low]=FLAG(R1,G1,B1);
		    else {
			d0=(double)(upp-low);
			for (idx=low; idx<=upp; idx++) {
			    d1=(double)(upp-idx)/d0;
			    d2=(double)(idx-low)/d0;
			    flag[idx]=FLAG((double)R1*d1+(double)R2*d2+0.5,
					   (double)G1*d1+(double)G2*d2+0.5,
					   (double)B1*d1+(double)B2*d2+0.5);
			}
		    }
		    break;
		default:
		    Error(bcm);
	    }
}
	(void)fprintf(stderr,"%d\t%d\t%d\n",Nx,Ny,Nz);

	if ((image=(Cell ***)malloc(sizeof(Cell **)*Nz))==NULL ||
	    (*image=(Cell **)malloc(sizeof(Cell *)*Nz*Ny))==NULL ||
	    (**image=(Cell *)calloc((size_t)Nz*Ny*Nx,sizeof(Cell)))==NULL ||
	    (lo=(U_S **)malloc(sizeof(U_S *)*(Ny+2)))==NULL ||
	    (*lo=(U_S *)calloc((size_t)(Ny+2)*(Nx+2),sizeof(U_S)))==NULL ||
	    (up=(U_S **)malloc(sizeof(U_S *)*(Ny+2)))==NULL ||
	    (*up=(U_S *)calloc((size_t)(Ny+2)*(Nx+2),sizeof(U_S)))==NULL)
	    Error("image memory allocation error.");

	++lo; ++lo[-1]; ++up; ++up[-1];
	for (y=0; y<=Ny; y++) {
	    lo[y]=lo[y-1]+Nx+2; up[y]=up[y-1]+Nx+2;
        }
	x1=Nx-1; y1=Ny-1; z1=Nz-1; x2=y2=z2=0;
	for (z=0; z<=Nz; z++) {
	    lu=lo; lo=up; up=lu;

	    if (z<Nz) {
		ReadImageFile(path[z],&nx,&ny,NULL,&cell,NULL); free(path[z]);

		if (z==0) {
		    slice=image[0]; y=1;
		}
		else {
		    slice=image[z]=image[z-1]+Ny; y=0;
		}
		for (; y<Ny; y++) slice[y]=slice[y-1]+Nx;

		for (y=0; y<ny; y++) {
		    for (x=0; x<nx; x++) slice[y][x]=cell[y][x];

		    free(cell[y]);
		}
		free(cell);

		for (y=0; y<Ny; y++)
		for (x=0; x<Nx; x++)
		    if ((lo[y][x]=flag[slice[y][x]])!=0) {
			if (x<x1) x1=x; if (x>x2) x2=x;
			if (y<y1) y1=y; if (y>y2) y2=y;
			if (z<z1) z1=z; if (z>z2) z2=z; ++pixels;
		    }
	    }
	    else {
		free(path);

		for (y=0; y<Ny; y++)
		for (x=0; x<Nx; x++) lo[y][x]=0;
	    }
	    for (y=0; y<=Ny; y++)
	    for (x=0; x<=Nx; x++)
{
	U_S	LNW=lo[y-1][x-1],LNE=lo[y-1][x],LSW=lo[y][x-1],LSE=lo[y][x],
		UNW=up[y-1][x-1],UNE=up[y-1][x],USW=up[y][x-1],USE=up[y][x];
	int	lnw=LNW!=0,lne=LNE!=0,lsw=LSW!=0,lse=LSE!=0,
		unw=UNW!=0,une=UNE!=0,usw=USW!=0,use=USE!=0,
		fx=(lnw!=lne)+(lsw!=lse)+(unw!=une)+(usw!=use),
		fy=(lnw!=lsw)+(lne!=lse)+(unw!=usw)+(une!=use),
		fz=(lnw!=unw)+(lne!=une)+(lsw!=usw)+(lse!=use),f=fx+fy+fz;

	if (f) {
	    if ((fx || LNW!=LNE || LSW!=LSE || UNW!=UNE || USW!=USE) &&
		(fy || LNW!=LSW || LNE!=LSE || UNW!=USW || UNE!=USE) &&
		(fz || LNW!=UNW || LNE!=UNE || LSW!=USW || LSE!=USE))
		++Vt;
	    else
		if (f==8) ++Vr;

	    ++Vs; faces+=f;
	}
}
	}
	free(up[-1]-1); free(up-1); free(lo[-1]-1); free(lo-1);

	if (pixels==0) Error("no object.");

	faces/=4; facets=2*(Vt-Vr-Vs+faces);
	(void)fprintf(stderr,"%lu\t%d\t%d\t%d\t%d\t%d\t%d\n%lu\t%lu\n",
			     pixels,x1,y1,z1,x2,y2,z2,faces,facets);

	nx=x2-x1+1;
	ny=y2-y1+1;
	nz=z2-z1+1; N=(nx>ny)?(nx>nz)?nx:nz:(ny>nz)?ny:nz;

	if ((lo=(U_S **)malloc(sizeof(U_S *)*(N+1)))==NULL ||
	    (*lo=(U_S *)calloc((size_t)(N+1)*(N+1),sizeof(U_S)))==NULL ||
	    (up=(U_S **)malloc(sizeof(U_S *)*(N+1)))==NULL ||
	    (*up=(U_S *)calloc((size_t)(N+1)*(N+1),sizeof(U_S)))==NULL ||
	    (AC0=(char *)calloc((size_t)(N+1),sizeof(char)))==NULL ||
	    (BD0=(char *)calloc((size_t)(N+1),sizeof(char)))==NULL ||
	    (AC1=(U_S *)calloc((size_t)(N+1),sizeof(U_S)))==NULL ||
	    (BD1=(U_S *)calloc((size_t)(N+1),sizeof(U_S)))==NULL ||
	    (AC2=(U_S *)calloc((size_t)(N*2+1),sizeof(U_S)))==NULL ||
	    (BD2=(U_S *)calloc((size_t)(N*2+1),sizeof(U_S)))==NULL ||
	    (vert=(List **)malloc(sizeof(List *)*(N*2+1)))==NULL ||
	    (help=(List **)malloc(sizeof(List *)*(N*2)))==NULL)
	    Error("work memory allocation error.");

	++lo; ++lo[-1]; ++up; ++up[-1];
	for (n=0; n<N; n++) {
	    lo[n]=lo[n-1]+N+1; up[n]=up[n-1]+N+1;
	}
	++AC0; ++BD0; ++AC1; ++BD1; ++AC2; ++BD2;

	if (argv[3][0]=='-')
	    file=stdout;
	else
	    if ((file=fopen(argv[3],"wb"))==NULL) Error("STL file not open.");

	for (n=0; n<80; n++)
	    if (fputc(0,file)!=0) Error("STL header write error.");
{
	U_S		u_s=0x0123;
        U_C		*u_c=(U_C *)&u_s;
        unsigned	u=(unsigned)facets;

	FWrite=(*u_c==(u_s&255))?(size_t (*)())fwrite:FWriteReverse;

	if ((*FWrite)((U_C *)&u,sizeof(unsigned),1,file)!=1)
	    Error("STL facet number write error.");
}
X=x1; Y=y1;

    Print=(void (*)())MinusZ;
/*
	for (y=0; y<ny; y++) for (x=0; x<nx; x++) lo[y][x]=0;
*/
	for (Z=z1; Z<=z2; Z++) {
	    lu=up; up=lo; lo=lu;

	    for (y0=Y, y=0; y<ny; y++, y0++)
	    for (x0=X, x=0; x<nx; x++, x0++) lo[y][x]=flag[image[Z][y0][x0]];

	    Erosion(nx,ny,lo,up);
	}

    Print=(void (*)())PlusZ;

	for (y=0; y<ny; y++) for (x=0; x<nx; x++) lo[y][x]=0;

	for (Z=z2; Z>=z1; Z--) {
	    lu=up; up=lo; lo=lu;

	    for (y0=Y, y=0; y<ny; y++, y0++)
	    for (x0=X ,x=0; x<nx; x++, x0++) lo[y][x]=flag[image[Z][y0][x0]];

	    Erosion(nx,ny,lo,up);
	}

Y=y1; Z=z1;

    Print=(void (*)())MinusX;

	for (z=0; z<nz; z++) for (y=0; y<ny; y++) lo[z][y]=0;

	for (X=x1; X<=x2; X++) {
	    lu=up; up=lo; lo=lu;

	    for (z0=Z, z=0; z<nz; z++, z0++)
	    for (y0=Y, y=0; y<ny; y++, y0++) lo[z][y]=flag[image[z0][y0][X]];

	    Erosion(ny,nz,lo,up);
	}

    Print=(void (*)())PlusX;

	for (z=0; z<nz; z++) for (y=0; y<ny; y++) lo[z][y]=0;

	for (X=x2; X>=x1; X--) {
	    lu=up; up=lo; lo=lu;

	    for (z0=Z, z=0; z<nz; z++, z0++)
	    for (y0=Y, y=0; y<ny; y++, y0++) lo[z][y]=flag[image[z0][y0][X]];

	    Erosion(ny,nz,lo,up);
	}

Z=z1; X=x1;

    Print=(void (*)())MinusY;

	for (x=0; x<nx; x++) for (z=0; z<nz; z++) lo[x][z]=0;

	for (Y=y1; Y<=y2; Y++) {
	    lu=up; up=lo; lo=lu;

	    for (x0=X, x=0; x<nx; x++, x0++)
	    for (z0=Z, z=0; z<nz; z++, z0++) lo[x][z]=flag[image[z0][Y][x0]];

	    Erosion(nz,nx,lo,up);
	}

    Print=(void (*)())PlusY;

	for (x=0; x<nx; x++) for (z=0; z<nz; z++) lo[x][z]=0;

	for (Y=y2; Y>=y1; Y--) {
	    lu=up; up=lo; lo=lu;

	    for (x0=X, x=0; x<nx; x++, x0++)
	    for (z0=Z, z=0; z<nz; z++, z0++) lo[x][z]=flag[image[z0][Y][x0]];

	    Erosion(nz,nx,lo,up);
	}
	if (file!=stdout) fclose(file);

#ifdef	DEBUG
	(void)fprintf(stderr,"%lu\n",stl_facets);
#endif

	free(help); free(vert);
	free(BD2-1); free(AC2-1); free(BD1-1); free(AC1-1);
	free(BD0-1); free(AC0-1);
	free(up[-1]-1); free(up-1); free(lo[-1]-1); free(lo-1);
	free(**image); free(*image); free(image);

	return 0;
}
