
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include "cell.h"
#include "rif.h"
#include "csi.h"

extern void	Error(char *msg);

#ifndef	P_R_I
#define P_R_I	1024
#endif

#define U_C	unsigned char
#define U_S	unsigned short
#define U_L	unsigned long

#define F_T	int
#define F_F	"%d"

#define RL_T	F_T
#include "rl.h"

typedef struct {
		U_S	x,y,z;
		U_L	index;
	} Point;
/*
static int	CompareZ(Point *p1,Point *p2)
{
	return (p1->z<p2->z)?-1:p1->z>p2->z;
}
*/
static int	CompareX(Point *p1,Point *p2)
{
	return (p1->x<p2->x)?-1:p1->x>p2->x;
}

static int	CompareY(Point *p1,Point *p2)
{
	return (p1->y<p2->y)?-1:p1->y>p2->y;
}

static FILE	*file;

static void	Open(char *path,char *vf,U_L vfs)
{
	if ((file=fopen(path,"w"))==NULL) {
	    (void)fputs(vf,stderr); Error(" file not open.");
	}
	(void)fprintf(file,"%lu\n",vfs);
}

typedef struct List {
		U_S		x,y;
		struct List	*next;
	} List;

static U_L	**PI;

static void	Plus(List *a,List *b,List *c,F_T flag)
{
	(void)fprintf(file,"%lu\t%lu\t%lu\t",
		      PI[a->y/2][a->x/2],PI[b->y/2][b->x/2],PI[c->y/2][c->x/2]);

	if (flag<0)
	    (void)fprintf(file,F_F "\n",~flag);
	else {
	    --flag; (void)fprintf(file,F_F "\t" F_F "\t" F_F "\n",
				       flag>>16,(flag>>8)&255,flag&255);
	}
}

static void	Minus(List *a,List *b,List *c,F_T flag)
{
	Plus(c,b,a,flag);
}

static void	(*Print)();

static void	Flush(List *list,F_T flag)
{
	U_L	idx,cnt=0;
	List	**array,**head,**tail,*oList,*nList,*stack=list;
	int	oSide,nSide,x0,y0,x1,y1,x2,y2;

	while ((list=list->next)!=NULL) ++cnt;

	if ((array=(List **)malloc(sizeof(List *)*cnt))==NULL)
	    Error("array memory allocation error.");

	list=stack;
	for (idx=0; idx<cnt; idx++) {
	    array[idx]=list; list=list->next;
	}
	free(list);

	stack->next=NULL;

	head=array+1; tail=array+cnt-1;
	if ((*head)->y>(*tail)->y) {
	    oSide=0; oList=(*(head++));
	}
	else {
	    oSide=1; oList=(*(tail--));
	}
	while (head!=tail) {
	    if ((*head)->y>(*tail)->y) {
		nSide=0; nList=(*(head++));
	    }
	    else {
		nSide=1; nList=(*(tail--));
	    }
	    if (oSide==nSide) {
		x2=oList->x-(x0=nList->x); y2=oList->y-(y0=nList->y);
		if (oSide==0)
		    do {
			x1=x2; y1=y2;
			if (y1*(x2=stack->x-x0)<=x1*(y2=stack->y-y0)) break;

			(*Print)(stack,oList,nList,flag); oList=stack;
		    } while ((stack=stack->next)!=NULL);
		else /* if (oSide==1) */
		    do {
			x1=x2; y1=y2;
			if (y1*(x2=stack->x-x0)>=x1*(y2=stack->y-y0)) break;

			(*Print)(stack,nList,oList,flag); oList=stack;
		    } while ((stack=stack->next)!=NULL);
	    }
	    else {
		list=oList;
		if (oSide==0)
		    do {
			(*Print)(stack,oList,nList,flag); oList=stack;
		    } while ((stack=stack->next)!=NULL);
		else /* if (oSide==1) */
		    do {
			(*Print)(stack,nList,oList,flag); oList=stack;
		    } while ((stack=stack->next)!=NULL);

		oList=list;
	    }
	    oList->next=stack; stack=oList; oList=nList; oSide=nSide;
	}
	nList=(*head);
	if (oSide==0)
	    do {
		(*Print)(stack,oList,nList,flag); oList=stack;
	    } while ((stack=stack->next)!=NULL);
	else /* if (oSide==1) */
	    do {
		(*Print)(stack,nList,oList,flag); oList=stack;
	    } while ((stack=stack->next)!=NULL);

	for (idx=0; idx<cnt; idx++) free(array[idx]); free(array);
}

static List	*Alloc(int x,int y,List *next)
{
	List	*list=(List *)malloc(sizeof(*list));

	if (list==NULL) Error("list memory allocation error.");

	list->x=(U_S)x; list->y=(U_S)y; list->next=next; return list;
}

static List	*hori,**vert,**help,**ptr;

static void	Corner(int x,int y,F_T a,F_T b,F_T c,F_T d,int h,int v)
{
	List	*temp,*next;

#define VERT	vert[x]
#define HELP	help[x]

	switch(((a!=0)?8:0)|((b!=0)?4:0)|((c!=0)?2:0)|((d!=0)?1:0)) {
	    case 0x1:
		VERT=Alloc(x,y+1,Alloc(x,y,hori=Alloc(x+1,y,NULL)));
		ptr=(&HELP);
		break;
	    case 0x2:
		VERT->y=(U_S)y;
		next=HELP->next;
		if (HELP->x==next->x && HELP->y==next->y) {
		    hori=Alloc(x+1,y,Alloc(x,y,next));
		    HELP->next=Alloc(x,y,NULL); Flush(VERT,c);
		}
		else
		    hori=Alloc(x+1,y,VERT);
		break;
	    case 0x3:
		if (v) {
		    VERT->y=(U_S)y; VERT=Alloc(x,y+1,VERT);
		}
		ptr=(&HELP);
		break;
	    case 0x4:
		hori->x=(U_S)x; hori->next=VERT=Alloc(x,y+1,NULL);
		*ptr=hori;
		break;
	    case 0x5:
		if (h) {
		    hori->x=(U_S)x; hori=hori->next=Alloc(x+1,y,NULL);
		}
		break;
/*
	    case 0x6:
		break;
*/
	    case 0x7:
		hori->x=(U_S)x; VERT->y=(U_S)y;
		next=HELP->next;
		if (HELP->x==next->x && HELP->y==next->y) {
		    hori->next=Alloc(x,y,next);
		    HELP->next=Alloc(x,y,NULL); Flush(VERT,d);
		}
		else
		    hori->next=VERT;
		*ptr=hori;
		break;
	    case 0x8:
		VERT->y=(U_S)y; hori->x=(U_S)x; Flush(hori,a);
		break;
/*
	    case 0x9:
		break;
*/
	    case 0xa:
		if (h) {
		    hori->x=(U_S)x; hori=Alloc(x+1,y,hori);
		}
		break;
	    case 0xb:
		hori->x=(U_S)x; VERT=Alloc(x,y+1,hori);
		ptr=(&HELP); HELP=hori;
		break;
	    case 0xc:
		if (v) {
		    VERT->y=(U_S)y; VERT=VERT->next=Alloc(x,y+1,NULL);
		}
		break;
	    case 0xd:
		VERT->y=(U_S)y;
		temp=(*ptr); next=temp->next;
		if (temp->x==next->x && temp->y==next->y) {
		    Flush(Alloc(x,y,next),b);
		    temp->next=Alloc(x,y,hori=Alloc(x+1,y,NULL));
		}
		else
		    VERT->next=hori=Alloc(x+1,y,NULL);
		break;
	    case 0xe:
		temp=(*ptr); next=temp->next;
		if (temp->x==next->x && temp->y==next->y)
		    hori=Alloc(x+1,y,Alloc(x,y,next));
		else
		    hori=Alloc(x+1,y,Alloc(x,y,Alloc(temp->x,temp->y,next)));
		*ptr=temp->next=Alloc(x,y,VERT=Alloc(x,y+1,NULL));
		break;
	}
}

static char	*AC0,*BD0;
static F_T	*AC1,*BD1,*AC2,*BD2;

static void	Erosion(int Nx,int Ny,F_T **lo,F_T **up)
{
	int	x1,x2,y1,y2,h,v,a,b,c,d;
	F_T	*lAC,*lBD,*uAC,*uBD,A,B,C,D;

	for (x1=x2=0; x1<Nx; x1++, x2+=2) {
	    BD0[x1]=0; BD1[x1]=BD2[x2]=BD2[x2+1]=0;
	}
	for (y1=y2=0; y1<Ny; y1++, y2+=2) {
	    lAC=lo[y1-1]; lBD=lo[y1];
	    uAC=up[y1-1]; uBD=up[y1];

	    for (x1=x2=0; x1<Nx; x1++, x2+=2) {
		AC0[x1]=BD0[x1]; BD0[x1]=(lBD[x1]!=0)|(uBD[x1]!=0)<<1;

		A=AC1[x1-1]; C=AC1[x1]=BD1[x1];
		B=BD1[x1-1]; D=BD1[x1]=(BD0[x1]==1)?lBD[x1]:0;

		if (D!=0 && D==C) {
		    AC2[x2]=(D==A && D==B)?D:0; AC2[x2+1]=D;
		}
		else
		    AC2[x2]=AC2[x2+1]=0;

		if (D!=B && D!=C)
		    h=v=1;
		else
		    if ((d=BD0[x1])==1)
			h=v=0;
		    else {
			b=BD0[x1-1]; c=AC0[x1];
			if (d==0) {
			    h=b==2 || b==3;
			    v=c==2 || c==3;
			}
			else { /* d==2 || d==3 */
			    h=b==0 || b+d==5 || (b==d && uBD[x1-1]!=uBD[x1]);
			    v=c==0 || c+d==5 || (c==d && uAC[x1  ]!=uBD[x1]);
			}
		    }

		Corner(x2,y2,BD2[x2-1],AC2[x2-1],BD2[x2],AC2[x2],h,v);

		Corner(x2+1,y2,BD2[x2],AC2[x2],BD2[x2+1],AC2[x2+1],0,0);
	    }
	    Corner(x2,y2,BD2[x2-1],AC2[x2-1],0,0,0,0);

	    for (x1=x2=0; x1<Nx; x1++, x2+=2) {
		A=AC1[x1-1]; C=AC1[x1];
		B=BD1[x1-1]; D=BD1[x1];

		if (D!=0) {
		    BD2[x2]=(D==B)?D:0; BD2[x2+1]=D;
		}
		else
		    BD2[x2]=BD2[x2+1]=0;

		Corner(x2,y2+1,AC2[x2-1],BD2[x2-1],AC2[x2],BD2[x2],0,0);

		if (A!=B && A!=C)
		    h=v=1;
		else
		    if ((a=AC0[x1-1])==1)
			h=v=0;
		    else {
			c=AC0[x1]; b=BD0[x1-1];
			if (a==0) {
			    h=c==2 || c==3;
			    v=b==2 || b==3;
			}
			else { /* a==2 || a==3 */
			    h=c==0 || c+a==5 || (c==a && uAC[x1  ]!=uAC[x1-1]);
			    v=b==0 || b+a==5 || (b==a && uBD[x1-1]!=uAC[x1-1]);
			}
		    }

		Corner(x2+1,y2+1,AC2[x2],BD2[x2],AC2[x2+1],BD2[x2+1],h,v);
            }
	    Corner(x2,y2+1,AC2[x2-1],BD2[x2-1],0,0,0,0);
	}
	for (x1=x2=0; x1<Nx; x1++, x2+=2) {
	    Corner(x2  ,y2,BD2[x2-1],0,BD2[x2  ],0,0,0);
	    Corner(x2+1,y2,BD2[x2  ],0,BD2[x2+1],0,0,0);
	}
        Corner(x2,y2,BD2[x2-1],0,0,0,0,0);
}

#define LIM	(size_t)((Cell)(~0))
#define LEN	1024

int	main(int argc,char **argv)
{
	char	**path;
	int	Nx,Ny,Nz,x,y,z,x1,y1,z1,x2,y2,z2,nx,ny,nz,N,n,x0,y0,z0;
	F_T	flag[LIM+1],**lo,**up,**lu;
	Cell	***image,**slice,**cell;
	U_L	pixels=0,points=0,Vr=0,Vs=0,faces=0;
	Point	*P,*point=NULL;

	if (argc!=3 && argc!=5)
	    Error("usage : si_vf_m directory nameFile {v.txt f.txt}");

	path=CheckSliceImages(argv[1],argv[2],&Nx,&Ny,&Nz,NULL);
{
	U_L	pv;
	char	str[LEN],rl[LEN];
	F_T	R,G,B;

	flag[0]=0; for (pv=1; pv<=LIM; pv++) flag[pv]=256*256*256;

	while (fgets(str,LEN,stdin)!=NULL && str[strlen(str)-1]=='\n')
	    switch(sscanf(str,"%s" " " F_F " " F_F " " F_F,rl,&R,&G,&B)) {
		case EOF:
		    continue;
		case 1:
		    RangeList(rl,LIM,flag,0); break;
		case 2:
		    if (R<0) Error("bad facet label.");

		    RangeList(rl,LIM,flag,~R); break;
		case 4:
		    if (R<0 || R>255 ||
			G<0 || G>255 ||
			B<0 || B>255) Error("bad facet color.");

		    RangeList(rl,LIM,flag,(R<<16)+(G<<8)+B+1); break;
		default:
		    Error("bad facet label and/or color.");
	    }
}
	if ((image=(Cell ***)malloc(sizeof(Cell **)*Nz))==NULL ||
	    (*image=(Cell **)malloc(sizeof(Cell *)*Nz*Ny))==NULL ||
	    (**image=(Cell *)calloc((size_t)Nz*Ny*Nx,sizeof(Cell)))==NULL ||
	    (lo=(F_T **)malloc(sizeof(F_T *)*(Ny+2)))==NULL ||
	    (*lo=(F_T *)calloc((size_t)(Ny+2)*(Nx+2),sizeof(F_T)))==NULL ||
	    (up=(F_T **)malloc(sizeof(F_T *)*(Ny+2)))==NULL ||
	    (*up=(F_T *)calloc((size_t)(Ny+2)*(Nx+2),sizeof(F_T)))==NULL)
	    Error("image memory allocation error.");

	++lo; ++lo[-1]; ++up; ++up[-1];
	for (y=0; y<=Ny; y++) {
	    lo[y]=lo[y-1]+Nx+2; up[y]=up[y-1]+Nx+2;
        }
	x1=Nx-1; y1=Ny-1; z1=Nz-1; x2=y2=z2=0;
	for (z=0; z<=Nz; z++) {
	    lu=lo; lo=up; up=lu;

	    if (z<Nz) {
		ReadImageFile(path[z],&nx,&ny,NULL,&cell,NULL); free(path[z]);

		if (z==0) {
		    slice=image[0]; y=1;
		}
		else {
		    slice=image[z]=image[z-1]+Ny; y=0;
		}
		for (; y<Ny; y++) slice[y]=slice[y-1]+Nx;

		for (y=0; y<ny; y++) {
		    for (x=0; x<nx; x++) slice[y][x]=cell[y][x];

		    free(cell[y]);
		}
		free(cell);

		for (y=0; y<Ny; y++)
		for (x=0; x<Nx; x++)
		    if ((lo[y][x]=flag[slice[y][x]])!=0) {
			if (x<x1) x1=x; if (x>x2) x2=x;
			if (y<y1) y1=y; if (y>y2) y2=y;
			if (z<z1) z1=z; if (z>z2) z2=z; ++pixels;
		    }
	    }
	    else {
		free(path);

		for (y=0; y<Ny; y++)
		for (x=0; x<Nx; x++) lo[y][x]=0;
	    }
	    for (y=0; y<=Ny; y++)
	    for (x=0; x<=Nx; x++)
{
	F_T	LNW=lo[y-1][x-1],LNE=lo[y-1][x],LSW=lo[y][x-1],LSE=lo[y][x],
		UNW=up[y-1][x-1],UNE=up[y-1][x],USW=up[y][x-1],USE=up[y][x];
	int	lnw=LNW!=0,lne=LNE!=0,lsw=LSW!=0,lse=LSE!=0,
		unw=UNW!=0,une=UNE!=0,usw=USW!=0,use=USE!=0,
		fx=(lnw!=lne)+(lsw!=lse)+(unw!=une)+(usw!=use),
		fy=(lnw!=lsw)+(lne!=lse)+(unw!=usw)+(une!=use),
		fz=(lnw!=unw)+(lne!=une)+(lsw!=usw)+(lse!=use),f=fx+fy+fz;

	if (f) {
	    if ((fx || LNW!=LNE || LSW!=LSE || UNW!=UNE || USW!=USE) &&
		(fy || LNW!=LSW || LNE!=LSE || UNW!=USW || UNE!=USE) &&
		(fz || LNW!=UNW || LNE!=UNE || LSW!=USW || LSE!=USE)) {
		if (points%P_R_I==0 &&
	    (point=(Point *)realloc(point,sizeof(Point)*(points+P_R_I)))==NULL)
		    Error("point memory allocation error.");

		P=point+points++; P->x=(U_S)x; P->y=(U_S)y; P->z=(U_S)z;
	    }
	    else
		if (f==8) ++Vr;

	    ++Vs; faces+=f;
	}
}
	}
	free(up[-1]-1); free(up-1); free(lo[-1]-1); free(lo-1);

	if (pixels==0) Error("no object.");

	nx=x2-x1+1;
	ny=y2-y1+1;
	nz=z2-z1+1; N=(nx>ny)?(nx>nz)?nx:nz:(ny>nz)?ny:nz;

	if ((points%P_R_I!=0 &&
	     (point=(Point *)realloc(point,sizeof(Point)*points))==NULL) ||
	    (PI=(U_L **)malloc(sizeof(U_L *)*(N+1)))==NULL ||
	    (*PI=(U_L *)malloc(sizeof(U_L)*(N+1)*(N+1)))==NULL ||
	    (lo=(F_T **)malloc(sizeof(F_T *)*(N+1)))==NULL ||
	    (*lo=(F_T *)calloc((size_t)(N+1)*(N+1),sizeof(F_T)))==NULL ||
	    (up=(F_T **)malloc(sizeof(F_T *)*(N+1)))==NULL ||
	    (*up=(F_T *)calloc((size_t)(N+1)*(N+1),sizeof(F_T)))==NULL ||
	    (AC0=(char *)calloc((size_t)(N+1),sizeof(char)))==NULL ||
	    (BD0=(char *)calloc((size_t)(N+1),sizeof(char)))==NULL ||
	    (AC1=(F_T *)calloc((size_t)(N+1),sizeof(F_T)))==NULL ||
	    (BD1=(F_T *)calloc((size_t)(N+1),sizeof(F_T)))==NULL ||
	    (AC2=(F_T *)calloc((size_t)(N*2+1),sizeof(F_T)))==NULL ||
	    (BD2=(F_T *)calloc((size_t)(N*2+1),sizeof(F_T)))==NULL ||
	    (vert=(List **)malloc(sizeof(List *)*(N*2+1)))==NULL ||
	    (help=(List **)malloc(sizeof(List *)*(N*2)))==NULL)
	    Error("work memory allocation error.");

	for (n=1; n<=N; n++) PI[n]=PI[n-1]+N+1;

	++lo; ++lo[-1]; ++up; ++up[-1];
	for (n=0; n<N; n++) {
	    lo[n]=lo[n-1]+N+1; up[n]=up[n-1]+N+1;
	}
	++AC0; ++BD0; ++AC1; ++BD1; ++AC2; ++BD2;
{
	U_L	i,facets=2*(points-Vr-Vs)+faces/2;

	if (argc==5)
	    Open(argv[3],"vertex",points);
	else {
	    file=stdout; (void)fprintf(file,"%lu\t%lu\n",points,facets);
	}
	for (P=point, i=0; i<points; i++, P++) {
	    (void)fprintf(file,"%hu\t%hu\t%hu\n",P->x,P->y,P->z);

	    P->x-=(U_S)x1; P->y-=(U_S)y1; P->z-=(U_S)z1; P->index=i;
	}
	if (file!=stdout) (void)fclose(file);

	if (argc==5) Open(argv[4],"facet",facets);
}
/*
qsort(point,(size_t)points,sizeof(Point),(int (*)())CompareZ);
*/
    P=point; Print=(void (*)())Minus;
/*
	for (y=0; y<ny; y++) for (x=0; x<nx; x++) lo[y][x]=0;
*/
	for (z0=z1, z=0; z<nz; z++, z0++) {
	    while (P->z==z) {
		PI[P->y][P->x]=P->index; ++P;
	    }
	    lu=up; up=lo; lo=lu;

	    for (y0=y1, y=0; y<ny; y++, y0++)
	    for (x0=x1, x=0; x<nx; x++, x0++) lo[y][x]=flag[image[z0][y0][x0]];

	    Erosion(nx,ny,lo,up);
	}

    P=point+points-1; Print=(void (*)())Plus;

	for (y=0; y<ny; y++) for (x=0; x<nx; x++) lo[y][x]=0;

	for (z0=z2, z=nz; z>0; z--, z0--) {
	    while (P->z==z) {
		PI[P->y][P->x]=P->index; --P;
	    }
	    lu=up; up=lo; lo=lu;

	    for (y0=y1, y=0; y<ny; y++, y0++)
	    for (x0=x1 ,x=0; x<nx; x++, x0++) lo[y][x]=flag[image[z0][y0][x0]];

	    Erosion(nx,ny,lo,up);
	}

qsort(point,(size_t)points,sizeof(Point),(int (*)())CompareX);

    P=point; Print=(void (*)())Minus;

	for (z=0; z<nz; z++) for (y=0; y<ny; y++) lo[z][y]=0;

	for (x0=x1, x=0; x<nx; x++, x0++) {
	    while (P->x==x) {
		PI[P->z][P->y]=P->index; ++P;
	    }
	    lu=up; up=lo; lo=lu;

	    for (z0=z1, z=0; z<nz; z++, z0++)
	    for (y0=y1, y=0; y<ny; y++, y0++) lo[z][y]=flag[image[z0][y0][x0]];

	    Erosion(ny,nz,lo,up);
	}

    P=point+points-1; Print=(void (*)())Plus;

	for (z=0; z<nz; z++) for (y=0; y<ny; y++) lo[z][y]=0;

	for (x0=x2, x=nx; x>0; x--, x0--) {
	    while (P->x==x) {
		PI[P->z][P->y]=P->index; --P;
	    }
	    lu=up; up=lo; lo=lu;

	    for (z0=z1, z=0; z<nz; z++, z0++)
	    for (y0=y1, y=0; y<ny; y++, y0++) lo[z][y]=flag[image[z0][y0][x0]];

	    Erosion(ny,nz,lo,up);
	}

qsort(point,(size_t)points,sizeof(Point),(int (*)())CompareY);

    P=point; Print=(void (*)())Minus;

	for (x=0; x<nx; x++) for (z=0; z<nz; z++) lo[x][z]=0;

	for (y0=y1, y=0; y<ny; y++, y0++) {
	    while (P->y==y) {
		PI[P->x][P->z]=P->index; ++P;
	    }
	    lu=up; up=lo; lo=lu;

	    for (x0=x1, x=0; x<nx; x++, x0++)
	    for (z0=z1, z=0; z<nz; z++, z0++) lo[x][z]=flag[image[z0][y0][x0]];

	    Erosion(nz,nx,lo,up);
	}

    P=point+points-1; Print=(void (*)())Plus;

	for (x=0; x<nx; x++) for (z=0; z<nz; z++) lo[x][z]=0;

	for (y0=y2, y=ny; y>0; y--, y0--) {
	    while (P->y==y) {
		PI[P->x][P->z]=P->index; --P;
	    }
	    lu=up; up=lo; lo=lu;

	    for (x0=x1, x=0; x<nx; x++, x0++)
	    for (z0=z1, z=0; z<nz; z++, z0++) lo[x][z]=flag[image[z0][y0][x0]];

	    Erosion(nz,nx,lo,up);
	}
	if (file!=stdout) (void)fclose(file);

	free(help); free(vert);
	free(BD2-1); free(AC2-1); free(BD1-1); free(AC1-1);
	free(BD0-1); free(AC0-1);
	free(up[-1]-1); free(up-1); free(lo[-1]-1); free(lo-1);
	free(*PI); free(PI); free(point);
	free(**image); free(*image); free(image);

	return 0;
}
