
#include <stdio.h>
#include <stdlib.h>
#include "cell.h"
#include "rif.h"
#include "csi.h"

extern void	Error(char *msg),
		RangeList(char *str,size_t lim,char *pv);

#ifndef	P_R_I
#define P_R_I	1024
#endif

#define U_C	unsigned char
#define U_S	unsigned short
#define U_L	unsigned long

typedef struct {
		U_S	x,y,z;
		U_L	index;
	} Point;
/*
static int	CompareZ(Point *p1,Point *p2)
{
	return (p1->z<p2->z)?-1:p1->z>p2->z;
}
*/
static int	CompareX(Point *p1,Point *p2)
{
	return (p1->x<p2->x)?-1:p1->x>p2->x;
}

static int	CompareY(Point *p1,Point *p2)
{
	return (p1->y<p2->y)?-1:p1->y>p2->y;
}

static FILE	*file;

static void	Open(char *path,char *vf,U_L vfs)
{
	if ((file=fopen(path,"w"))==NULL) {
	    (void)fputs(vf,stderr); Error(" file not open.");
	}
	(void)fprintf(file,"%lu\n",vfs);
}

typedef struct List {
		U_S		x,y;
		struct List	*next;
	} List;

static U_L	**PI;

static void	Plus(List *a,List *b,List *c)
{
	(void)fprintf(file,"%lu\t%lu\t%lu\n",
			   PI[a->y][a->x],PI[b->y][b->x],PI[c->y][c->x]);
}

static void	Minus(List *a,List *b,List *c)
{
	Plus(c,b,a);
}

static void	(*Print)();

static void	Flush(List *list)
{
	U_L	idx,cnt=0;
	List	**array,**head,**tail,*oList,*nList,*stack=list;
	int	oSide,nSide,x0,y0,x1,y1,x2,y2;

	while ((list=list->next)!=NULL) ++cnt;

	if ((array=(List **)malloc(sizeof(List *)*cnt))==NULL)
	    Error("array memory allocation error.");

	list=stack;
	for (idx=0; idx<cnt; idx++) {
	    array[idx]=list; list=list->next;
	}
	free(list);

	stack->next=NULL;

	head=array+1; tail=array+cnt-1;
	if ((*head)->y>(*tail)->y) {
	    oSide=0; oList=(*(head++));
	}
	else {
	    oSide=1; oList=(*(tail--));
	}
	while (head!=tail) {
	    if ((*head)->y>(*tail)->y) {
		nSide=0; nList=(*(head++));
	    }
	    else {
		nSide=1; nList=(*(tail--));
	    }
	    if (oSide==nSide) {
		x2=oList->x-(x0=nList->x); y2=oList->y-(y0=nList->y);
		if (oSide==0)
		    do {
			x1=x2; y1=y2;
			if (y1*(x2=stack->x-x0)<=x1*(y2=stack->y-y0)) break;

			(*Print)(stack,oList,nList); oList=stack;
		    } while ((stack=stack->next)!=NULL);
		else /* if (oSide==1) */
		    do {
			x1=x2; y1=y2;
			if (y1*(x2=stack->x-x0)>=x1*(y2=stack->y-y0)) break;

			(*Print)(stack,nList,oList); oList=stack;
		    } while ((stack=stack->next)!=NULL);
	    }
	    else {
		list=oList;
		if (oSide==0)
		    do {
			(*Print)(stack,oList,nList); oList=stack;
		    } while ((stack=stack->next)!=NULL);
		else /* if (oSide==1) */
		    do {
			(*Print)(stack,nList,oList); oList=stack;
		    } while ((stack=stack->next)!=NULL);

		oList=list;
	    }
	    oList->next=stack; stack=oList; oList=nList; oSide=nSide;
	}
	nList=(*head);
	if (oSide==0)
	    do {
		(*Print)(stack,oList,nList); oList=stack;
	    } while ((stack=stack->next)!=NULL);
	else /* if (oSide==1) */
	    do {
		(*Print)(stack,nList,oList); oList=stack;
	    } while ((stack=stack->next)!=NULL);

	for (idx=0; idx<cnt; idx++) free(array[idx]); free(array);
}

static List	*Alloc(int x,int y,List *next)
{
	List	*list=(List *)malloc(sizeof(*list));

	if (list==NULL) Error("list memory allocation error.");

	list->x=(U_S)x; list->y=(U_S)y; list->next=next; return list;
}

static List	*hori,**vert,**help,**ptr;

static void	Corner(int x,int y,int a,int b,int c,int d)
{
	List	*temp,*next;

#define VERT	vert[x]
#define HELP	help[x]

	switch(((a==1)?8:0)|((b==1)?4:0)|((c==1)?2:0)|((d==1)?1:0)) {
	    case 0x1:
		VERT=Alloc(x,y+1,Alloc(x,y,hori=Alloc(x+1,y,NULL)));
		ptr=(&HELP);
		break;
	    case 0x2:
		VERT->y=(U_S)y;
		next=HELP->next;
		if (HELP->x==next->x && HELP->y==next->y) {
		    hori=Alloc(x+1,y,Alloc(x,y,next));
		    HELP->next=Alloc(x,y,NULL); Flush(VERT);
		}
		else
		    hori=Alloc(x+1,y,VERT);
		break;
	    case 0x3:
		if (a!=b) {
		    VERT->y=(U_S)y; VERT=Alloc(x,y+1,VERT);
		}
		ptr=(&HELP);
		break;
	    case 0x4:
		hori->x=(U_S)x; hori->next=VERT=Alloc(x,y+1,NULL);
		*ptr=hori;
		break;
	    case 0x5:
		if (a!=c) {
		    hori->x=(U_S)x; hori=hori->next=Alloc(x+1,y,NULL);
		}
		break;
	    case 0x6:
		VERT->y=(U_S)y;
		next=HELP->next;
		if (HELP->x==next->x && HELP->y==next->y) {
		    temp=Alloc(x+1,y,Alloc(x,y,next));
		    HELP->next=Alloc(x,y,NULL); Flush(VERT);
		}
		else
		    temp=Alloc(x+1,y,VERT);
		hori->x=(U_S)x; hori->next=VERT=Alloc(x,y+1,NULL);
		*ptr=hori;
		hori=temp;
		break;
	    case 0x7:
		hori->x=(U_S)x; VERT->y=(U_S)y;
		next=HELP->next;
		if (HELP->x==next->x && HELP->y==next->y) {
		    hori->next=Alloc(x,y,next);
		    HELP->next=Alloc(x,y,NULL); Flush(VERT);
		}
		else
		    hori->next=VERT;
		*ptr=hori;
		break;
	    case 0x8:
		VERT->y=(U_S)y; hori->x=(U_S)x; Flush(hori);
		break;
	    case 0x9:
		VERT->y=(U_S)y; hori->x=(U_S)x; Flush(hori);
		VERT=Alloc(x,y+1,Alloc(x,y,hori=Alloc(x+1,y,NULL)));
		ptr=(&HELP);
		break;
	    case 0xa:
		if (b!=d) {
		    hori->x=(U_S)x; hori=Alloc(x+1,y,hori);
		}
		break;
	    case 0xb:
		hori->x=(U_S)x; VERT=Alloc(x,y+1,hori);
		ptr=(&HELP); HELP=hori;
		break;
	    case 0xc:
		if (c!=d) {
		    VERT->y=(U_S)y; VERT=VERT->next=Alloc(x,y+1,NULL);
		}
		break;
	    case 0xd:
		VERT->y=(U_S)y;
		temp=(*ptr); next=temp->next;
		if (temp->x==next->x && temp->y==next->y) {
		    Flush(Alloc(x,y,next));
		    temp->next=Alloc(x,y,hori=Alloc(x+1,y,NULL));
		}
		else
		    VERT->next=hori=Alloc(x+1,y,NULL);
		break;
	    case 0xe:
		temp=(*ptr); next=temp->next;
		if (temp->x==next->x && temp->y==next->y)
		    hori=Alloc(x+1,y,Alloc(x,y,next));
		else
		    hori=Alloc(x+1,y,Alloc(x,y,Alloc(temp->x,temp->y,next)));
		*ptr=temp->next=Alloc(x,y,VERT=Alloc(x,y+1,NULL));
		break;
	}
}

#define LIM	(size_t)((Cell)(~0))

int	main(int argc,char **argv)
{
	char	**path,pv[LIM+1],**LU;
	int	Nx,Ny,Nz,n,y,x1,y1,z1,x2,y2,z2,z,nx,ny,x,nz,N,x0,y0,z0;
	U_C	***image,**slice,mask;
	Cell	**cell;
	U_L	pixels=0,points=0,Vr=0,Vs=0,faces=0;
	Point	*P,*point=NULL;

	if (argc!=4 && argc!=6)
	    Error("usage : si_vf_s directory nameFile rangeList {v.txt f.txt}");

	path=CheckSliceImages(argv[1],argv[2],&Nx,&Ny,&Nz,NULL);

	RangeList(argv[3],LIM,pv);

	n=(Nz+7)/8;
	if ((image=(U_C ***)malloc(sizeof(U_C **)*n))==NULL ||
	    (*image=(U_C **)malloc(sizeof(U_C *)*n*Ny))==NULL ||
	    (**image=(U_C *)calloc((size_t)n*Ny*Nx,sizeof(U_C)))==NULL ||
	    (LU=(char **)malloc(sizeof(char *)*(Ny+2)))==NULL ||
	    (*LU=(char *)calloc((size_t)(Ny+2)*(Nx+2),sizeof(char)))==NULL)
	    Error("image memory allocation error.");

	++LU; ++LU[-1]; for (y=0; y<=Ny; y++) LU[y]=LU[y-1]+Nx+2;

	x1=Nx-1; y1=Ny-1; z1=Nz-1; x2=y2=z2=0;
	for (z=0; z<=Nz; z++) {
	    if (z<Nz) {
		ReadImageFile(path[z],&nx,&ny,NULL,&cell,NULL); free(path[z]);

		if (z%8==0) {
		    if (z==0) {
			slice=image[0]; y=1;
		    }
		    else {
			n=z/8; slice=image[n]=image[n-1]+Ny; y=0;
		    }
		    for (; y<Ny; y++) slice[y]=slice[y-1]+Nx; mask=1;
		}
		else
		    mask<<=1;

		for (y=0; y<Ny; y++) {
		    for (x=0; x<Nx; x++)
			if (pv[(x<nx && y<ny)?cell[y][x]:0]) {
			    slice[y][x]|=mask;
			    if (x<x1) x1=x; if (x>x2) x2=x;
			    if (y<y1) y1=y; if (y>y2) y2=y;
			    if (z<z1) z1=z; if (z>z2) z2=z; ++pixels;

			    LU[y][x]=((LU[y][x]<<1)|1)&3;
			}
			else
			    LU[y][x]=(LU[y][x]<<1)&2;

		    if (y<ny) free(cell[y]);
		}
		free(cell);
	    }
	    else {
		free(path);

		for (y=0; y<Ny; y++)
		for (x=0; x<Nx; x++) LU[y][x]=(LU[y][x]<<1)&2;
            }
	    for (y=0; y<=Ny; y++)
	    for (x=0; x<=Nx; x++)
{
	int	lnw=LU[y-1][x-1]&1,unw=LU[y-1][x-1]>>1,
		lne=LU[y-1][x  ]&1,une=LU[y-1][x  ]>>1,
		lsw=LU[y  ][x-1]&1,usw=LU[y  ][x-1]>>1,
		lse=LU[y  ][x  ]&1,use=LU[y  ][x  ]>>1,
		fx=(lnw!=lne)+(lsw!=lse)+(unw!=une)+(usw!=use),
		fy=(lnw!=lsw)+(lne!=lse)+(unw!=usw)+(une!=use),
		fz=(lnw!=unw)+(lne!=une)+(lsw!=usw)+(lse!=use),f=fx+fy+fz;

	if (f) {
	    if (fx && fy && fz) {
		if (points%P_R_I==0 &&
	    (point=(Point *)realloc(point,sizeof(Point)*(points+P_R_I)))==NULL)
		    Error("point memory allocation error.");

		P=point+points++; P->x=(U_S)x; P->y=(U_S)y; P->z=(U_S)z;
	    }
	    else
		if (f==8) ++Vr;

	    ++Vs; faces+=f;
	}
}
	}
	free(LU[-1]-1); free(LU-1);

	if (pixels==0) Error("no object.");

	nx=x2-x1+1;
	ny=y2-y1+1;
	nz=z2-z1+1; N=(nx>ny)?(nx>nz)?nx:nz:(ny>nz)?ny:nz;

	n=N+1;
	if ((points%P_R_I!=0 &&
	     (point=(Point *)realloc(point,sizeof(Point)*points))==NULL) ||
	    (PI=(U_L **)malloc(sizeof(U_L *)*n))==NULL ||
	    (*PI=(U_L *)malloc(sizeof(U_L)*n*n))==NULL ||
	    (LU=(char **)malloc(sizeof(char *)*n))==NULL ||
	    (*LU=(char *)calloc((size_t)n*n,sizeof(char)))==NULL ||
	    (vert=(List **)malloc(sizeof(List *)*n))==NULL ||
	    (help=(List **)malloc(sizeof(List *)*N))==NULL)
	    Error("work memory allocation error.");

	for (n=1; n<=N; n++) PI[n]=PI[n-1]+N+1;

	++LU; ++LU[-1]; for (n=0; n<N; n++) LU[n]=LU[n-1]+N+1;
{
	U_L	i,facets=2*(points-Vr-Vs)+faces/2;

	if (argc==6)
	    Open(argv[4],"vertex",points);
	else {
	    file=stdout; (void)fprintf(file,"%lu\t%lu\n",points,facets);
	}
	for (P=point, i=0; i<points; i++, P++) {
	    (void)fprintf(file,"%hu\t%hu\t%hu\n",P->x,P->y,P->z);

	    P->x-=(U_S)x1; P->y-=(U_S)y1; P->z-=(U_S)z1; P->index=i;
	}
	if (file!=stdout) (void)fclose(file);

	if (argc==6) Open(argv[5],"facet",facets);
}
#define CORNER(h,v)	Corner(h,v,LU[v-1][h-1],LU[v][h-1],LU[v-1][h],LU[v][h])
#define CORNER_H(h,v)	Corner(h,v,LU[v-1][h-1],LU[v][h-1],0,0)
#define CORNER_V(h,v)	Corner(h,v,LU[v-1][h-1],0,LU[v-1][h],0)
#define CORNER_HV(h,v)	Corner(h,v,LU[v-1][h-1],0,0,0)
/*
qsort(point,(size_t)points,sizeof(Point),(int (*)())CompareZ);
*/
    P=point; Print=(void (*)())Minus;
/*
	for (y=0; y<ny; y++) for (x=0; x<nx; x++) LU[y][x]=0;
*/
	for (z0=z1, z=0; z<nz; z++, z0++) {
	    while (P->z==z) {
		PI[P->y][P->x]=P->index; ++P;
	    }
	    slice=image[z0/8]; mask=(U_C)(1<<(z0%8));
	    for (y0=y1, y=0; y<ny; y++, y0++) {
		for (x0=x1, x=0; x<nx; x++, x0++) {
		    LU[y][x]=((LU[y][x]<<1)|((slice[y0][x0]&mask)!=0))&3;

		    CORNER(x,y);
		}
		CORNER_H(x,y);
	    }
	    for (x=0; x<nx; x++) CORNER_V(x,y);

	    CORNER_HV(x,y);
	}

    P=point+points-1; Print=(void (*)())Plus;

	for (y=0; y<ny; y++) for (x=0; x<nx; x++) LU[y][x]=0;

	for (z0=z2, z=nz; z>0; z--, z0--) {
	    while (P->z==z) {
		PI[P->y][P->x]=P->index; --P;
	    }
	    slice=image[z0/8]; mask=(U_C)(1<<(z0%8));
	    for (y0=y1, y=0; y<ny; y++, y0++) {
		for (x0=x1, x=0; x<nx; x++, x0++) {
		    LU[y][x]=((LU[y][x]<<1)|((slice[y0][x0]&mask)!=0))&3;

		    CORNER(x,y);
		}
		CORNER_H(x,y);
	    }
	    for (x=0; x<nx; x++) CORNER_V(x,y);

	    CORNER_HV(x,y);
	}

qsort(point,(size_t)points,sizeof(Point),(int (*)())CompareX);

    P=point; Print=(void (*)())Minus;

	for (z=0; z<nz; z++) for (y=0; y<ny; y++) LU[z][y]=0;

	for (x0=x1, x=0; x<nx; x++, x0++) {
	    while (P->x==x) {
		PI[P->z][P->y]=P->index; ++P;
	    }
	    for (z0=z1, z=0; z<nz; z++, z0++) {
		slice=image[z0/8]; mask=(U_C)(1<<(z0%8));
		for (y0=y1, y=0; y<ny; y++, y0++) {
		    LU[z][y]=((LU[z][y]<<1)|((slice[y0][x0]&mask)!=0))&3;

		    CORNER(y,z);
		}
		CORNER_H(y,z);
	    }
	    for (y=0; y<ny; y++) CORNER_V(y,z);

	    CORNER_HV(y,z);
	}

    P=point+points-1; Print=(void (*)())Plus;

	for (z=0; z<nz; z++) for (y=0; y<ny; y++) LU[z][y]=0;

	for (x0=x2, x=nx; x>0; x--, x0--) {
	    while (P->x==x) {
		PI[P->z][P->y]=P->index; --P;
	    }
	    for (z0=z1, z=0; z<nz; z++, z0++) {
		slice=image[z0/8]; mask=(U_C)(1<<(z0%8));
		for (y0=y1, y=0; y<ny; y++, y0++) {
		    LU[z][y]=((LU[z][y]<<1)|((slice[y0][x0]&mask)!=0))&3;

		    CORNER(y,z);
		}
		CORNER_H(y,z);
	    }
	    for (y=0; y<ny; y++) CORNER_V(y,z);

	    CORNER_HV(y,z);
	}

qsort(point,(size_t)points,sizeof(Point),(int (*)())CompareY);

    P=point; Print=(void (*)())Minus;

	for (x=0; x<nx; x++) for (z=0; z<nz; z++) LU[x][z]=0;

	for (y0=y1, y=0; y<ny; y++, y0++) {
	    while (P->y==y) {
		PI[P->x][P->z]=P->index; ++P;
	    }
	    for (x0=x1, x=0; x<nx; x++, x0++) {
		for (z0=z1, z=0; z<nz; z++, z0++) {
	    LU[x][z]=((LU[x][z]<<1)|((image[z0/8][y0][x0]&(1<<(z0%8)))!=0))&3;

		    CORNER(z,x);
		}
		CORNER_H(z,x);
	    }
	    for (z=0; z<nz; z++) CORNER_V(z,x);

	    CORNER_HV(z,x);
	}

    P=point+points-1; Print=(void (*)())Plus;

	for (x=0; x<nx; x++) for (z=0; z<nz; z++) LU[x][z]=0;

	for (y0=y2, y=ny; y>0; y--, y0--) {
	    while (P->y==y) {
		PI[P->x][P->z]=P->index; --P;
	    }
	    for (x0=x1, x=0; x<nx; x++, x0++) {
		for (z0=z1, z=0; z<nz; z++, z0++) {
	    LU[x][z]=((LU[x][z]<<1)|((image[z0/8][y0][x0]&(1<<(z0%8)))!=0))&3;

		    CORNER(z,x);
		}
		CORNER_H(z,x);
	    }
	    for (z=0; z<nz; z++) CORNER_V(z,x);

	    CORNER_HV(z,x);
	}
	if (file!=stdout) (void)fclose(file);

	free(help); free(vert); free(LU[-1]-1); free(LU-1);
	free(*PI); free(PI); free(point);
	free(**image); free(*image); free(image);

	return 0;
}
